package be.iminds.ilabt.jfed.lowlevel.api_wrapper;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AbacCredential;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.SfaCredential;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.ListIterator;

/**
 * ApiWrapper
 */
public abstract class ApiWrapper {
    private final static org.slf4j.Logger parentLOG = LoggerFactory.getLogger(ApiWrapper.class);

    protected final Logger logger;
    protected final GeniUserProvider geniUserProvider;
    protected final JFedConnectionProvider connectionProvider;
    protected final JFedPreferences jFedPreferences;

    /** If this is not empty and not null, speaksFor is activated */
    protected List<AnyCredential> speaksForCredentials;
    /** Optional URN of the user spoken for. May be null. */
    protected GeniUrn speaksForUserUrn;

    protected ApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        assert logger != null;
        assert geniUserProvider != null;
        assert connectionProvider != null;
        assert jFedPreferences != null;

        this.logger = logger;
        this.geniUserProvider = geniUserProvider;
        this.connectionProvider = connectionProvider;
        this.jFedPreferences = jFedPreferences;
    }

    /**
     * returns only the SfaCredentials in the list, discarding the ABAC and other credentials, unless they are speaksfor.
     * */
    public List<AnyCredential> filterCredentials(List<AnyCredential> origCredentials) {
        if (origCredentials == null)
            return null;
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        for (AnyCredential cred : origCredentials) {
            if (cred instanceof SfaCredential) {
                res.add(cred);
            }
            if (cred instanceof AbacCredential && cred.isSpeaksFor()) {
                res.add(cred);
            }
        }
        if (res.isEmpty() && !origCredentials.isEmpty()) {
            parentLOG.warn("Unknown what is a good credential to use. Will just use first one received (type=" + origCredentials.get(0).getClass().getName() + ").");
            res.add(origCredentials.get(0));
        }

        parentLOG.debug("ApiWrapper.filterCredentials started with  "+origCredentials.size()+" credentials and ended with "+res.size());
        return res;
    }

    /*
     * activate speaksfor support for this Wrapper.
     *
     * NOTE: speaksFor credential itself can be supplied in 2 ways:
     *    - as part of argument list of credentials each option takes
     *    - using this option
     * these wrappers will prevent the exact same credential being sent twice.
     *    -> see addSpeaksForCredentials
     */
    public void setSpeaksFor(List<AnyCredential> speaksForCredentials, @Nullable GeniUrn speakingForUrn) {
        this.speaksForCredentials = speaksForCredentials == null ? Collections.<AnyCredential>emptyList() : speaksForCredentials;
        this.speaksForUserUrn = speakingForUrn;
    }
    public boolean isSpeaksFor() {
        return this.speaksForCredentials != null && !this.speaksForCredentials.isEmpty();
    }
    public List<AnyCredential> getSpeaksForCredentials() {
        return speaksForCredentials;
    }
    public GeniUrn getSpeaksForUserUrn() {
        return speaksForUserUrn;
    }

    /**
     * Adds speaks_for credentials if needed and not yet added.
     * */
    public List<AnyCredential> addSpeaksForCredentials(List<AnyCredential> origCredentials) {
        if (origCredentials == null)
            return null;
        List<AnyCredential> res = new ArrayList<AnyCredential>(origCredentials);

        List<AnyCredential> speaksForCredsToAdd = speaksForCredentials == null ? null : new ArrayList<AnyCredential>(speaksForCredentials);
        if (speaksForCredsToAdd != null && !speaksForCredsToAdd.isEmpty()) {
            for (AnyCredential cred : origCredentials) {
                for (ListIterator<AnyCredential> it = speaksForCredsToAdd.listIterator(); it.hasNext(); ) {
                    AnyCredential speaksForCredential = it.next();
                    if (speaksForCredential.getCredentialXml().equals(cred.getCredentialXml())) {
                        //handle duplicate: don't add it
                        it.remove();
                    }
                }
            }
            res.addAll(speaksForCredsToAdd);
        }
        parentLOG.debug("AggregateManagerWrapper.addSpeaksForCredentials started with  "+origCredentials.size()+" credentials and ended with "+res.size());
        return res;
    }

    /**
     * returns only the SfaCredentials in the list, discarding the ABAC and other credentials, unless they are speaksfor.
     * Also adds speaks_for credentials if needed and not yet added.
     *
     * this is the combination of filterCredentials and addSpeaksForCredentials
     * */
    public List<AnyCredential> finalizeSliceCredentials(List<AnyCredential> origCredentials) {
        if (origCredentials == null)
            return null;
        List<AnyCredential> res = filterCredentials(origCredentials);
        res = addSpeaksForCredentials(res);
        if (res != null)
            res = new ArrayList<AnyCredential>(res);

        parentLOG.debug("AggregateManagerWrapper.finalizeSliceCredentials started with  "+origCredentials.size()+" credentials and ended with "+res.size());
        return res;
    }

}
